/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_FRAME_DESCRIPTOR_HXX_
# define PCAP_FRAME_DESCRIPTOR_HXX_

# include <cstring>

# include <boost/lexical_cast.hpp>

# include "frame_descriptor.hh"
# include <wipal/tool/exceptions.hh>
# include <wipal/tool/endianness.hh>

namespace pcapxx
{

  template <class PCAPDescriptor>
  frame_descriptor<PCAPDescriptor>::frame_descriptor(const pkthdr_ptr&	    h,
						     const bytes_ptr&	    b,
						     unsigned		    i,
						     const PCAPDescriptor*  d):
    data_ (new shared_data (d, i, h, b))
  {
  }

  template <class PCAPDescriptor>
  template <class D>
  frame_descriptor<PCAPDescriptor>::
  frame_descriptor(const descriptor<D>&	desc,
		   std::istream&	stream,
		   unsigned		idx,
		   std::streampos*	pos)
  {
    pkthdr_ptr	pcap_header;
    bytes_ptr	bytes;

    // PCAP header.
    {
      internals::file_frame_header h;

      if (not stream.read(reinterpret_cast<char*> (&h), sizeof (h)) or
	  stream.gcount() != sizeof (h))
	throw tool::read_error ("Unreadable packet header");
      if (pos)
	*pos += sizeof (h);

      const bool		s = desc.swapped();
      pkthdr* const		p = new pkthdr;

      // Clear padding.  Allows further use of memcmp() to test
      // equality between PCAP headers.
      memset(p, 0, sizeof (*p));

      p->ts.tv_sec	= tool::extract_long_s(h.ts.tv_sec, s) + desc.zone();
      p->ts.tv_usec	= tool::extract_long_s(h.ts.tv_usec, s);
      p->caplen		= tool::extract_long_u(h.caplen, s);
      p->len		= tool::extract_long_u(h.len, s);
      pcap_header.reset(p);
    }

    // Frame body.
    {
      const size_t	caplen = pcap_header->caplen;

      if (caplen > internals::max_packet_size)
	throw tool::bad_file("Packet size too large to make sense (" +
			     boost::lexical_cast<std::string>(caplen) + ")");

      uint8_t* const	tmp = new uint8_t[caplen];

      if (not stream.read(reinterpret_cast<char*> (tmp), caplen) or
	  stream.gcount() != std::streampos (caplen))
	{
	  delete tmp;
	  throw tool::read_error ("Unreadable packet data");
	}
      if (pos)
	*pos += caplen;
      bytes.reset(tmp);
    }

    data_.reset(new shared_data (desc.exact_ptr(), idx, pcap_header, bytes));
  }

  namespace internals
  {

    template <class D1, class D2>
    struct fd_ctor_helper
    {
      typedef typename frame_descriptor<D1>::shared_data	shared_data1;
      typedef typename frame_descriptor<D2>::shared_data	shared_data2;
      typedef boost::shared_ptr<shared_data1>			shared_ptr1;
      typedef boost::shared_ptr<shared_data2>			shared_ptr2;


      static void set(shared_ptr1& d1, const shared_ptr2& d2)
      {
	assert(d2);

	d1.reset(new shared_data1 (0, 0, d2->pcap_header, d2->bytes));
      }
    };

    template <class D>
    struct fd_ctor_helper<D, D>
    {
      typedef typename frame_descriptor<D>::shared_data		shared_data;
      typedef boost::shared_ptr<shared_data>			shared_ptr;

      static void set(shared_ptr& d1, const shared_ptr& d2)
      {
	d1 = d2;
      }
    };

  } // End of namespace internals.

  template <class PCAPDescriptor>
  template <class OtherDescriptor>
  frame_descriptor<PCAPDescriptor>::
  frame_descriptor(const frame_descriptor<OtherDescriptor>& fd)
  {
    typedef internals::fd_ctor_helper<PCAPDescriptor, OtherDescriptor> helper;

    helper::set(data_, fd.data());
  }

  template <class PCAPDescriptor>
  unsigned
  frame_descriptor<PCAPDescriptor>::id() const
  {
    assert(data_);

    return data_->id;
  }

  namespace internals
  {
    template <class D>
    bool
    fd_swapped_helper(const D*)
    {
      throw std::logic_error ("This code should be unreachable. "
			      "Please fill a bug report");
    }

    template <class D>
    bool
    fd_swapped_helper(const descriptor<D>* d)
    {
      return d ? d->swapped() : false;
    }

    template <>
    inline
    bool
    fd_swapped_helper<tool::undefined_type>(const tool::undefined_type* d)
    {
      assert(not d);
      (void) d; // Avoid unused parameter warnings when NDEBUG.

      return false;
    }

  } // End of namespace pcapxx::internals.

  template <class PCAPDescriptor>
  bool
  frame_descriptor<PCAPDescriptor>::swapped() const
  {
    return internals::fd_swapped_helper(desc_ptr());
  }

  template <class PCAPDescriptor>
  const PCAPDescriptor*
  frame_descriptor<PCAPDescriptor>::desc_ptr() const
  {
    assert(data_);

    return data_->desc;
  }

  template <class PCAPDescriptor>
  const PCAPDescriptor&
  frame_descriptor<PCAPDescriptor>::desc() const
  {
    assert(data_);

    if (not data_->desc)
      throw std::runtime_error ("Frame has no refering PCAP descriptor");
    return *data_->desc;
  }

  template <class PCAPDescriptor>
  const typename frame_descriptor<PCAPDescriptor>::pkthdr_ptr&
  frame_descriptor<PCAPDescriptor>::pcap_header() const
  {
    assert(data_);

    return data_->pcap_header;
  }

  template <class PCAPDescriptor>
  const typename frame_descriptor<PCAPDescriptor>::bytes_ptr&
  frame_descriptor<PCAPDescriptor>::bytes() const
  {
    assert(data_);

    return data_->bytes;
  }

  namespace internals
  {

    template <class D>
    std::ostream&
    fd_print_helper(std::ostream& o, const D*)
    {
      throw std::logic_error ("This code should be unreachable. "
			      "Please fill a bug report");
    }

    template <class D>
    std::ostream&
    fd_print_helper(std::ostream& o, const descriptor<D>* d)
    {
      if (d)
	o << d->file_name() << ':';
      return o;
    }

    template <>
    inline
    std::ostream&
    fd_print_helper<tool::undefined_type>
    (std::ostream& o, const tool::undefined_type* d)
    {
      assert(not d);
      (void) d; // Avoid unused parameter warnings when NDEBUG.

      return o;
    }

  } // End of namespace pcapxx::internals.

  template <class PCAPDescriptor>
  std::ostream&
  frame_descriptor<PCAPDescriptor>::print(std::ostream& os) const
  {
    return internals::fd_print_helper(os, desc_ptr()) << data_->id;
  }

  template <class PCAPDescriptor>
  bool
  frame_descriptor<PCAPDescriptor>::
  operator == (const frame_descriptor& rhs) const
  {
    return
      0 == memcmp(pcap_header().get(), rhs.pcap_header().get(), sizeof (pkthdr))
      and
      0 == memcmp(bytes().get(), rhs.bytes().get(), pcap_header()->caplen);
  }

  template <class PCAPDescriptor>
  bool
  frame_descriptor<PCAPDescriptor>::
  operator != (const frame_descriptor& rhs) const
  {
    return not (*this == rhs);
  }

  template <class PCAPDescriptor>
  frame_descriptor<PCAPDescriptor>::
  shared_data::shared_data(const PCAPDescriptor*	desc,
			   unsigned			id,
			   const pkthdr_ptr&		pcap_header,
			   const bytes_ptr&		bytes):
    desc (desc),
    id (id),
    pcap_header (pcap_header),
    bytes (bytes)
  {
  }

  template <class PCAPDescriptor>
  const
    boost::shared_ptr<typename frame_descriptor<PCAPDescriptor>::shared_data>&
  frame_descriptor<PCAPDescriptor>::data() const
  {
    assert(data_);

    return data_;
  }

  template <class PCAPDescriptor>
  std::ostream&
  operator << (std::ostream& os, const frame_descriptor<PCAPDescriptor>& fd)
  {
    return fd.print(os);
  }

} // End of namespace pcapxx.

#endif // ! PCAP_FRAME_DESCRIPTOR_HXX_
