/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_ADDR_HXX_
# define TOOL_ADDR_HXX_

# include <cstring>
# include <cassert>
# include <sstream>
# include <iomanip>
# include <stdexcept>
# include <sstream>

# include "addr.hh"

namespace tool
{

  template <class VT, unsigned N, bool HN, char SN>
  addr<VT, N, HN, SN>::addr()
  {
    memset(addr_, 0, sizeof (addr_));
  }

  template <class VT, unsigned N, bool HN, char SN>
  addr<VT, N, HN, SN>::addr(const void* a)
  {
    memcpy(addr_, a, sizeof (addr_));
  }

  template <class VT, unsigned N, bool HN, char SN>
  addr<VT, N, HN, SN>::addr(const std::string& str)
  {
    std::istringstream	is (str);

    read(is);
    if (not is)
      throw std::invalid_argument ("string is not a valid address");
  }

  template <class VT, unsigned N, bool HN, char SN>
  addr<VT, N, HN, SN>::addr(std::istream& is)
  {
    read(is);
    if (not is)
      throw std::invalid_argument ("string is not a valid address");
  }

  template <class VT, unsigned N, bool HN, char SN>
  std::istream&
  addr<VT, N, HN, SN>::read(std::istream& is)
  {
    if (HN)
      is >> std::hex;
    for (unsigned i = 0; i < N; ++i)
      {
	unsigned	t;

	if (i and is.get() != SN)
	  {
	    is.setstate(is.rdstate() | std::istream::failbit);
	    return is;
	  }
	if (not (is >> t))
	  return is;
	addr_[i] = VT (t);
      }
    return is;
  }

  template <class VT, unsigned N, bool HN, char SN>
  bool
  addr<VT, N, HN, SN>::operator < (const addr& rhs) const
  {
    return memcmp(addr_, rhs.addr_, sizeof (addr_)) < 0;
  }

  template <class VT, unsigned N, bool HN, char SN>
  bool
  addr<VT, N, HN, SN>::operator == (const addr& rhs) const
  {
    return memcmp(addr_, rhs.addr_, sizeof (addr_)) == 0;
  }

  template <class VT, unsigned N, bool HN, char SN>
  bool
  addr<VT, N, HN, SN>::operator != (const addr& rhs) const
  {
    return not operator == (rhs);
  }

  template <class VT, unsigned N, bool HN, char SN>
  VT&
  addr<VT, N, HN, SN>::operator [] (unsigned i)
  {
    assert(i < N);

    return addr_[i];
  }

  template <class VT, unsigned N, bool HN, char SN>
  const VT&
  addr<VT, N, HN, SN>::operator [] (unsigned i) const
  {
    assert(i < N);

    return addr_[i];
  }

  template <class VT, unsigned N, bool HN, char SN>
  void
  addr<VT, N, HN, SN>::dump(void *a) const
  {
    memcpy(a, addr_, sizeof (addr_));
  }

  template <class VT, unsigned N, bool HN, char SN>
  const addr<VT, N, HN, SN>&
  addr<VT, N, HN, SN>::null()
  {
    static const addr<VT, N, HN, SN>	null;

    return null;
  }

  template <class VT, unsigned N, bool HN, char SN>
  const addr<VT, N, HN, SN>&
  addr<VT, N, HN, SN>::broadcast()
  {
    static bool			init = false;
    static addr<VT, N, HN, SN>	bcast;

    if (not init)
      {
	memset(bcast.addr_, 0xFF, sizeof (bcast.addr_));
	init = true;
      }
    return bcast;
  }

  template <class VT, unsigned N, bool HN, char SN>
  std::ostream&
  operator << (std::ostream& os, const addr<VT, N, HN, SN>& a)
  {
    if (HN)
      os << std::hex << std::setfill ('0');
    for (unsigned i = 0; i < N - 1; ++i)
      {
	if (HN)
	  os << std::setw (sizeof (VT) * 2);
	os << unsigned (a[i]) << SN;
      }
    if (HN)
      os << std::setw (sizeof (VT) * 2);
    return os << unsigned (a[N - 1])
	      << std::dec << std::setfill (' ');
  }

  template <class VT, unsigned N, bool HN, char SN>
  std::istream&
  operator >> (std::istream& is, addr<VT, N, HN, SN>& a)
  {
    return a.read(is);
  }

  template <class VT, unsigned N, bool HN, char SN>
  std::string
  make_string(const addr<VT, N, HN, SN>& a)
  {
    std::ostringstream	os;

    os << a;
    return os.str();
  }

} // End of namespace tool.

#endif // ! TOOL_ADDR_HXX_
