/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_ARGS_HH_
# define TOOL_ARGS_HH_

# include <vector>
# include <string>
# include <set>
# include <utility>

# include <wipal/tool/options.hh>
# include <wipal/pcap/descriptor.hh>


/*--------.
| Options |
`--------*/

# define WP_ARGS_OPTIONS						\
    ((8, "Only compare 802.11 frames (no PHY header, no PCAP)."))	\
    ((a, ("Attributes for unique frames identification (default: " +	\
	  def::attributes() + ").")))					\
    ((b, "Only compare frame bytes (no PCAP header)."))			\
    ((c, "Do not print column headers (default if output is not a tty).")) \
    ((C, "Print column headers (default if output is a tty)."))		\
    ((d, "Compare everything: PCAP data and bytes (default)."))		\
    ((e, "Do not print error values (default)."))			\
    ((E, "Print error values."))					\
    ((g, "Print debug information."))					\
    ((h, "Print this help."))						\
    ((i, "Do not print frame indices."))				\
    ((I, "Print frame indices (default)."))				\
    ((m, ("MAC addresses mapping file (default: " + def::mac_map() + ")."))) \
    ((n, "Force little endian for PHY headers."))			\
    ((N, "Force big endian for PHY headers."))				\
    ((o, "Compare everything and count how many bytes differ."))	\
    ((p, "Prism noise fields have no special meaning (default)."))	\
    ((P, "Consider a non-zero Prism noise field as a PHY error."))	\
    ((q, "Quiet ; produce minimal output (no column headers, etc.)."))	\
    ((r, "Ignore reference.  Blacklist a given x-y reference frame."))	\
    ((s, ("ESSID mapping file (default: " + def::essid_map() + ").")))	\
    ((t, "Compare 802.11 frames and time of arrival. "			\
      "Use a precision of 106us."))					\
    ((u, "Do not print microseconds (default)."))			\
    ((U, "Print microseconds."))					\
    ((v, "Print version information."))					\
    ((x, "Precision when comparing timestamps (default: 106)."))	\
    ((y, "Use PCAP timestamps."))					\
    ((Y, "Use PHY header timestamps when available (default)."))


/*-------------.
| Attributes.  |
`-------------*/

# define WP_ATTR_HELP_hsh_en2		"802.3  - MD5 hash."
# define WP_ATTR_HELP_hsh_80211		"802.11 - MD5 hash (only parse MAC " \
					          "headers)."
# define WP_ATTR_HELP_hsh_80211_x	"802.11 - MD5 hash (decapsulate " \
						  "frames)."
# define WP_ATTR_HELP_tmp		"802.11 - Tstamp."
# define WP_ATTR_HELP_dst_tmp		"802.11 - dst., Tstamp."
# define WP_ATTR_HELP_src_tmp		"802.11 - src., Tstamp."
# define WP_ATTR_HELP_bss_tmp		"802.11 - BSS., Tstamp."
# define WP_ATTR_HELP_src_bss_tmp	"802.11 - src., BSS., Tstamp."
# define WP_ATTR_HELP_seq_tmp		"802.11 - SeqCtl, Tstamp."
# define WP_ATTR_HELP_seq_bss_tmp	"802.11 - SeqCtl, BSS., Tstamp."
# define WP_ATTR_HELP_seq_dst_bss_tmp	"802.11 - SeqCtl, dst., BSS., Tstamp."
# define WP_ATTR_HELP_seq_src_bss_tmp	"802.11 - SeqCtl, src., BSS., Tstamp."


namespace tool
{

  namespace args
  {
    typedef std::vector<std::string>			input;
    typedef std::vector<std::string>			output;

    enum comp_method
      {
	cmp_default,
	cmp_802,
	cmp_bytes,
	cmp_time,
	cmp_score
      };

    enum count
      {
	none		= 0,
	arbitrary	= 8,
      };

    void
    parse(int		argc,
	  char* const*	argv,
	  opt::list&	opts,
	  count		ninputs,
	  count		noutputs);

    template <class Fun>
    int
    dispatch_unique_id(opt::list& options, Fun fun);

    void
    unsupported_linktype(const pcapxx::descriptor<>::link_type lt);

    void
    check_options(opt::list& o, const pcapxx::descriptor<>::link_type l);

    void
    ensure_80211(const pcapxx::descriptor<>::link_type lt);

    /// Default values for program arguments.
    namespace def
    {

      endian::endianness			endianness();
      std::set< std::pair<unsigned, unsigned> >	ref_blacklist();
      undefined_type				mapping();
      bool					ignore_noisy_prism();
      std::string				mac_map();
      std::string				essid_map();
      std::string				attributes();
      comp_method				cmp();
      bool					ui_errors();
      bool					ui_indices();
      bool					ui_headers();
      bool					ui_microseconds();
      bool					ui_summary();

      /**
       * Default precision for timestamp comparisons.
       *
       * 106 microseconds.  This is largest value one can afford when
       * synchronizing IEEE 802.11 traces.
       *
       * @see	phy::eq_time_and_80211,
       *	Yeo et al., A framework for wireless LAN monitoring and
       *	its applications.
       */
      unsigned					precision();

      bool					phy_tstamps();

    } // End of namespace tool::args::def.

  } // End of namespace tool::args.

} // End of namespace tool.

# include "args.hxx"

#endif // TOOL_ARGS_HH_
