/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_ARGS_HXX_
# define TOOL_ARGS_HXX_

# include <iostream>
# include <stdexcept>
# include <cstdlib>
# include <cassert>
# include <sstream>

extern "C"
{
# include <unistd.h>
}

# include <boost/preprocessor/seq/for_each.hpp>
# include <boost/preprocessor/stringize.hpp>
# include <boost/preprocessor/cat.hpp>
# include <boost/preprocessor/tuple/elem.hpp>
# include <boost/preprocessor/control/if.hpp>
# include <boost/preprocessor/seq/for_each_i.hpp>

# include <boost/algorithm/string/split.hpp>
# include <boost/algorithm/string/classification.hpp>

# include <boost/lexical_cast.hpp>

# include <wipal/tool/args.hh>
# include <wipal/wifi/addr_mapping.hh>

# ifdef HAVE_CONFIG_H
#  include <config.hh>
# endif
# ifdef HAVE_OPENSSL
#  include <wipal/wifi/frame/unique_id/hash.hh>
# endif // HAVE_OPENSSL
# include <wipal/wifi/frame/unique_id/timestamp.hh>
# include <wipal/wifi/frame/unique_id/addr_timestamp.hh>
# include <wipal/wifi/frame/unique_id/source_bssid_timestamp.hh>
# include <wipal/wifi/frame/unique_id/seqctl_timestamp.hh>
# include <wipal/wifi/frame/unique_id/seqctl_bssid_timestamp.hh>
# include <wipal/wifi/frame/unique_id/seqctl_dest_bssid_timestamp.hh>
# include <wipal/wifi/frame/unique_id/seqctl_source_bssid_timestamp.hh>


namespace tool
{

  namespace args
  {

    namespace internals
    {

      /*-------------.
      | Help strings |
      `-------------*/

      struct help_string
      {
	help_string(const std::string& c,
		    const std::string& h): c_ (c), h_ (h)
	{
	}

	std::ostream&	print(std::ostream& os) const
	{
	  return os << "  -" << c_ << "   " << h_ << '\n';
	}

      private:
	std::string	c_;
	std::string	h_;
      };

      inline
      std::ostream&
      operator << (std::ostream& os, const help_string& hs)
      {
	return hs.print(os);
      }

# define WP_args_declare_help_string(R, Unused, Opt)		\
      extern							\
      const help_string						\
      BOOST_PP_CAT(help_, BOOST_PP_TUPLE_ELEM(2, 0, Opt));

      BOOST_PP_SEQ_FOR_EACH(WP_args_declare_help_string, ~, WP_ARGS_OPTIONS)

# undef WP_args_declare_help_string


      /*------.
      | Usage |
      `------*/

      inline
      void
      usage(const std::string&	progname,
	    const opt::list&	opts,
	    const count&	ninputs,
	    const count&	noutputs)
      {
	std::cerr << "Invocation:\n\t" << progname << " [options]";
	switch (ninputs)
	  {
	  case 0:					break;
	  case 1:  std::cerr << " <input>";		break;
	  case 2:  std::cerr << " <input-1> <input-2>";	break;
	  default: std::cerr << " <inputs...>";		break;
	  }
	switch (noutputs)
	  {
	  case 0:					break;
	  case 1:  std::cerr << " <output>";		break;
	  default: std::cerr << " <outputs...>";	break;
	  }

	std::cerr << "\n\nOptions:\n";
	if (opts.count("misc"))
	  {
	    const opt::list&  misc = opt::get<opt::list>(opts["misc"]);

	    if (misc.count("cmp"))			std::cerr << help_8
								  << help_b
								  << help_d
								  << help_o
								  << help_t;
	    if (misc.count("attributes"))		std::cerr << help_a;
	    if (misc.count("mac_map"))			std::cerr << help_m;
	    if (misc.count("essid_map"))		std::cerr << help_s;
	  }
	if (opts.count("internals"))
	  {
	    const opt::list&  internals = opt::get<opt::list>(opts
							      ["internals"]);

	    if (internals.count("endianness"))		std::cerr << help_n
								  << help_N;
	    if (internals.count("ignore_noisy_prism"))	std::cerr << help_p
								  << help_P;
	    if (internals.count("ref_blacklist"))	std::cerr << help_r;
	    if (internals.count("phy_tstamps"))		std::cerr << help_y
								  << help_Y;
	  }
	if (opts.count("ui"))
	  {
	    const opt::list&	ui = opt::get<opt::list>(opts["ui"]);

	    if (ui.count("headers"))			std::cerr << help_c
								  << help_C;
	    if (ui.count("errors"))			std::cerr << help_e
								  << help_E;
	    if (ui.count("indices"))			std::cerr << help_i
								  << help_I;
	    if (ui.count("microseconds"))		std::cerr << help_u
								  << help_U;
	    std::cerr << help_q;
	  }
	std::cerr << help_g << help_h << help_v;

# define WP_help_attr_string(R, Unused, A)	\
	  "  " BOOST_PP_STRINGIZE(A) ":\t"	\
	    BOOST_PP_CAT(WP_ATTR_HELP_, A) "\n"

	if (opts.count("misc") and
	    opt::get<opt::list>(opts["misc"]).count("attributes"))
	  std::cerr << "\nAvailable attributes:\n"
	    BOOST_PP_SEQ_FOR_EACH(WP_help_attr_string, ~, WP_ATTRIBUTES);

# undef WP_help_attr_string

# define WP_help_linktypes_string(R, Unused, LT)			\
	"  " BOOST_PP_STRINGIZE(BOOST_PP_TUPLE_ELEM(2, 0, LT)) "\n"

	std::cerr << "\nSupported PCAP link types:\n"
	  BOOST_PP_SEQ_FOR_EACH(WP_help_linktypes_string, ~, WP_LINKTYPES);

# undef WP_help_linktypes_string
      }


      /*--------.
      | Version |
      `--------*/

      inline
      void
      version(const std::string& progname)
      {
	std::cerr << progname << "\n"
	  "Trace Tools, version " << PACKAGE_VERSION << '.' << std::endl;
      }


      /*-----------------.
      | make_reference() |
      `-----------------*/

      inline
      std::pair<unsigned, unsigned>
      make_reference(const std::string& s)
      {
	static const std::string	err = ("Invalid reference frame "
					       "format. Format is "
					       "``<unsigned>-<unsigned>''");
	std::vector<std::string>	v;

	boost::split(v, s, boost::is_any_of("-"));
	if (v.size() != 2)
	  throw std::invalid_argument (err);
	try
	  {
	    return std::make_pair(boost::lexical_cast<unsigned> (v[0]),
				  boost::lexical_cast<unsigned> (v[1]));
	  }
	catch (const boost::bad_lexical_cast&)
	  {
	    throw std::invalid_argument (err);
	  }
      }

    } // End of namespace tool:args::internals.

    /*------------.
    | args::parse |
    `------------*/

    inline
    void
    parse(int		argc,
	  char* const*	argv,
	  opt::list&	opts,
	  count		ninputs,
	  count		noutputs)
    {
      using opt::get;

      opts.add((opt::name ("proc") -= opt::list
		((opt::name("input") -= input(),
		  opt::name("output") -= output())),
		opt::name ("internals") -= opt::list
		((opt::name ("debug") -= false))));

      int	o;

      try
	{
	  while ((o = getopt(argc, argv,
			     "8a:bcCdeEghiIm:nNpPqr:s:touUvx:yY")) != -1)
	    switch (o)
	      {
	      case '8':
		opts["misc"]["cmp"] = cmp_802;
		break;

	      case 'a':
		opts["misc"]["attributes"] = std::string (optarg);
		break;

	      case 'b':
		opts["misc"]["cmp"] = cmp_bytes;
		break;

	      case 'c':
		opts["ui"]["headers"] = false;
		break;

	      case 'C':
		opts["ui"]["headers"] = true;
		break;

	      case 'd':
		opts["misc"]["cmp"] = cmp_default;
		break;

	      case 'e':
		opts["ui"]["errors"] = false;
		break;

	      case 'E':
		opts["ui"]["errors"] = true;
		break;

	      case 'g':
		opts["internals"]["debug"] = true;
		break;

	      case 'h':
		internals::usage(argv[0], opts, ninputs, noutputs);
		exit(EXIT_SUCCESS);
		break;

	      case 'i':
		opts["ui"]["indices"] = false;
		break;

	      case 'I':
		opts["ui"]["indices"] = true;
		break;

	      case 'm':
		opts["misc"]["mac_map"] = std::string (optarg);
		break;

	      case 'n':
		opts["internals"]["endianness"] = endian::little;
		break;

	      case 'N':
		opts["internals"]["endianness"] = endian::big;
		break;

	      case 'o':
		opts["misc"]["cmp"] = cmp_score;
		break;

	      case 'p':
		opts["internals"]["ignore_noisy_prism"] = false;
		break;

	      case 'P':
		opts["internals"]["ignore_noisy_prism"] = true;
		break;

	      case 'q':
		{
		  opt::list&	ui = get<opt::list>(opts["ui"]);

		  if (ui.count("errors"))	ui["errors"] = false;
		  if (ui.count("indices"))	ui["indices"] = false;
		  if (ui.count("headers"))	ui["headers"] = false;
		  if (ui.count("microseconds"))	ui["microseconds"] = false;
		  if (ui.count("summary"))	ui["summary"] = false;
		}
		break;

	      case 'r':
		{
		  typedef std::set< std::pair<unsigned, unsigned> > blist_type;

		  get<blist_type>(opts["internals"]["ref_blacklist"]).
		    insert(internals::make_reference(optarg));
		}
		break;

	      case 's':
		opts["misc"]["essid_map"] = std::string (optarg);
		break;

	      case 't':
		opts["misc"]["cmp"] = cmp_time;
		break;

	      case 'u':
		opts["ui"]["microseconds"] = false;
		break;

	      case 'U':
		opts["ui"]["microseconds"] = true;
		break;

	      case 'v':
		internals::version(argv[0]);
		exit(EXIT_SUCCESS);

	      case 'x':
		opts["internals"]["precision"] =
		  boost::lexical_cast<unsigned>(optarg);
		break;

	      case 'y':
		opts["internals"]["phy_tstamps"] = false;
		break;

	      case 'Y':
		opts["internals"]["phy_tstamps"] = true;
		break;

	      default: /* '?' */
		internals::usage(argv[0], opts, ninputs, noutputs);
		exit(EXIT_FAILURE);
	      }
	}
      catch (const opt::invalid_option& e)
	{
	  std::cerr
	    << argv[0] << " invalid option -- " << char (o) << std::endl;
	  internals::usage(argv[0], opts, ninputs, noutputs);
	  exit(EXIT_FAILURE);
	}

      if (opts.count("internals") and
	  get<opt::list>(opts["internals"]).count("mapping"))
	{
	  const std::string n = get<std::string>(opts["misc"]["mac_map"]);

	  opts["internals"]["mapping"] = wifi::addr_mapping (n);
	}

      assert(ninputs != args::arbitrary or noutputs != args::arbitrary);

      if (ninputs == args::arbitrary)
	ninputs = args::count (argc - optind - noutputs);
      if (noutputs == args::arbitrary)
	noutputs = args::count (argc - optind - ninputs);
      if (ninputs < 0 or noutputs < 0 or optind + ninputs + noutputs != argc)
	{
	  std::cerr << argv[0] << ": wrong number of arguments." << std::endl;
	  internals::usage(argv[0], opts, ninputs, noutputs);
	  exit(EXIT_FAILURE);
	}

      input&	in = get<input>(opts["proc"]["input"]);
      output&	out = get<output>(opts["proc"]["output"]);

      for (int i = 0; i < ninputs; ++i)
	in.push_back(argv[optind + i]);
      for (int i = 0; i < noutputs; ++i)
	out.push_back(argv[optind + ninputs + i]);

      if (get<bool>(opts["internals"]["debug"]))
	std::cerr << "DEBUG: args::parse() finished.\n"
		  << "DEBUG: Options:\n" << opts;
    }


    /*-------------------.
    | dispatch_unique_id |
    `-------------------*/

    template <class Fun>
    int
    dispatch_unique_id(opt::list& options, Fun fun)
    {
      assert(options.count("misc"));
      assert(opt::get<opt::list>(options["misc"]).count("attributes"));

      const std::string&	attr = opt::get<std::string>(options["misc"]
							     ["attributes"]);

      if (false)
	;
# define WP_dispatch_unique_id_else(R, Unused, A)	\
      else if (attr == BOOST_PP_STRINGIZE(A))		\
	return fun.template operator ()			\
	  <wifi::frame::BOOST_PP_CAT(A, _id)>(options);

      BOOST_PP_SEQ_FOR_EACH(WP_dispatch_unique_id_else, ~, WP_ATTRIBUTES)

# undef WP_dispatch_unique_id_else

	throw std::invalid_argument ("Invalid distinctive attributes");
    }


    /*---------------------.
    | unsupported_linktype |
    `---------------------*/

    inline
    void
    unsupported_linktype(const pcapxx::descriptor<>::link_type lt)
    {
      std::ostringstream	os;

# define WP_linktype_string(R, Unused, I, LT)			\
      BOOST_PP_IF(I, ", ", "")					\
	BOOST_PP_STRINGIZE(BOOST_PP_TUPLE_ELEM(2, 0, LT))

      os << "Unsupported PCAP link type: " << lt << " - expected one of ["
	BOOST_PP_SEQ_FOR_EACH_I(WP_linktype_string, ~, WP_LINKTYPES) "]";

# undef WP_linktype_string

      throw std::invalid_argument (os.str());
    }

    /*--------------.
    | check_options |
    `--------------*/

    inline
    void
    check_options(opt::list& o, const pcapxx::descriptor<>::link_type l)
    {
      // If given '-P', verify link type is PRISM_HEADER.
      if (o.count("internals") and
	  o["internals"].get<opt::list>().count("ignore_noisy_prism") and
	  o["internals"]["ignore_noisy_prism"].get<bool>() and
	  l != pcapxx::descriptor<>::PRISM_HEADER)
	throw std::invalid_argument ("Expected PRISM_HEADER because of '-P', "
				     "got " +
				     boost::lexical_cast<std::string>(l));

      if (o.count("misc") and o["misc"].get<opt::list>().count("attributes"))
	{
	  const std::string&	attr = (o["misc"]["attributes"]
					.get<std::string>());

	  // If given "-a hsh_en2", verify link type is EN10MB...
	  if (attr == "hsh_en2")
	    {
	      if (l != pcapxx::descriptor<>::EN10MB)
		throw std::invalid_argument ("Expected EN10MB because of "
					     "\"-a hsh_en2\", got " +
					     boost::lexical_cast<std::string>
					     (l));
	    }
	  // ... Else, verify link type *is not* EN10MB.
	  else if (l == pcapxx::descriptor<>::EN10MB)
	    throw std::invalid_argument ("EN10MB traces only works with "
					 "\"-a hsh_en2\"");
	}
    }

    /*-------------.
    | ensure_80211 |
    `-------------*/

    void
    ensure_80211(const pcapxx::descriptor<>::link_type lt)
    {
      switch (lt)
	{
	case pcapxx::descriptor<>::IEEE802_11:
	case pcapxx::descriptor<>::IEEE802_11_RADIO:
	case pcapxx::descriptor<>::IEEE802_11_RADIO_AVS:
	case pcapxx::descriptor<>::PRISM_HEADER:
	  return;

	default:
	  throw std::invalid_argument ("Expected a link type that includes "
				       "IEEE 802.11 frames, got " +
				       boost::lexical_cast<std::string>(lt));
	}
    }

    namespace def
    {

      inline
      endian::endianness
      endianness()
      {
	return endian::guess;
      }

      inline
      std::set< std::pair<unsigned, unsigned> >
      ref_blacklist()
      {
	return std::set< std::pair<unsigned, unsigned> > ();
      }

      inline
      undefined_type
      mapping()
      {
	return undefined_type ();
      }

      inline
      bool
      ignore_noisy_prism()
      {
	return false;
      }

      inline
      std::string
      mac_map()
      {
	return "MAC.map";
      }

      inline
      std::string
      essid_map()
      {
	return "ESSID.map";
      }

      inline
      std::string
      attributes()
      {
	return BOOST_PP_STRINGIZE(WP_DEFAULT_ATTRIBUTES);
      }

      inline
      comp_method
      cmp()
      {
	return cmp_default;
      }

      inline
      bool
      ui_errors()
      {
	return false;
      }

      inline
      bool
      ui_indices()
      {
	return true;
      }

      inline
      bool
      ui_headers()
      {
	return isatty(STDOUT_FILENO);
      }

      inline
      bool
      ui_microseconds()
      {
	return false;
      }

      inline
      bool
      ui_summary()
      {
	return true;
      }

      inline
      unsigned
      precision()
      {
	return 106;
      }

      inline
      bool
      phy_tstamps()
      {
	return true;
      }

    } // End of namespace tool::args::def.

  } // End of namespace tool::opt::args.

} // End of namespace tool.

#endif // TOOL_ARGS_HXX_

