/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_ENDIANNESS_HH_
# define TOOL_ENDIANNESS_HH_

extern "C"
{
# include <inttypes.h>
}

# include <ostream>

namespace tool
{

  /// Endianness related compounds.
  namespace endian
  {
    /// Endianness tag.
    enum endianness
      {
	big,	///< Tag a data as big endian.
	little,	///< Tag a data as little endian.
	raw,	///< Specify the data does not need endianness adjustments.
	swap,	///< Specify the data needs to be swapped.
	guess	///< Unknown endianness, algorithm must make a guess.
      };

    std::ostream&
    operator << (std::ostream&, const endianness& e);

    /**
     * Return true when conversion from @c tag to the host byte order
     * needs a swap.
     *
     * If @c tag is @c endian::guess then return @c guess_swap.
     */
    bool	need_swap(endianness tag, bool guess_swap);
  }


  /*----------------.
  | Unsigned shorts |
  `----------------*/

  /// Consider @c p as a pointer to unsigned shorts and return *p.
  uint16_t
  extract_unswapped_short_u(const void* p);

  /// Return @c s with swapped low and high bytes.
  uint16_t
  extract_swapped_short_u(const uint16_t s);

  /// Consider @c p as a pointer to endian-swapped unsigned shorts and
  ///  return *p.
  uint16_t
  extract_swapped_short_u(const void* p);

  /// Return @c *p, possibly with endianness adjustment if @c swapped.
  uint16_t
  extract_short_u(const void* p, bool swapped);

  /// Return @c s, possibly with endianness adjustment if @c swapped.
  uint16_t
  extract_short_u(const uint16_t s, bool swapped);

  /// Consider @c p as a pointer to big-endian unsigned shorts and return *p.
  uint16_t
  extract_big_endian_short_u(const void* p);

  /// Swap @c s if needed (i.e. host is little endian).
  uint16_t
  extract_big_endian_short_u(const uint16_t s);

  /// Consider @c p as a pointer to little-endian unsigned shorts and
  /// return *p.
  uint16_t
  extract_little_endian_short_u(const void* p);

  /// Swap @c s if needed (i.e. host is big endian).
  uint16_t
  extract_little_endian_short_u(const uint16_t s);


  /*---------------.
  | Unsigned longs |
  `---------------*/

  /// Consider @c p as a pointer to unsigned longs and return *p.
  uint32_t
  extract_unswapped_long_u(const void* p);

  /// Return @c s with swapped low and high bytes.
  uint32_t
  extract_swapped_long_u(const uint32_t l);

  /// Consider @c p as a pointer to endian-swapped unsigned longs and
  ///  return *p.
  uint32_t
  extract_swapped_long_u(const void* p);

  /// Return @c *p, possibly with endianness adjustment if @c swapped.
  uint32_t
  extract_long_u(const void* p, bool swapped);

  /// Return @c s, possibly with endianness adjustment if @c swapped.
  uint32_t
  extract_long_u(const uint32_t l, bool swapped);

  /// Consider @c p as a pointer to big-endian unsigned longs and return *p.
  uint32_t
  extract_big_endian_long_u(const void* p);

  /// Swap @c s if needed (i.e. host is little endian).
  uint32_t
  extract_big_endian_long_u(const uint32_t l);

  /// Consider @c p as a pointer to little-endian unsigned longs and
  /// return *p.
  uint32_t
  extract_little_endian_long_u(const void* p);

  /// Swap @c s if needed (i.e. host is big endian).
  uint32_t
  extract_little_endian_long_u(const uint32_t l);


  /*-------------.
  | Signed longs |
  `-------------*/

  /// Consider @c p as a pointer to unsigned longs and return *p.
  int32_t
  extract_unswapped_long_s(const void* p);

  /// Return @c s with swapped low and high bytes.
  int32_t
  extract_swapped_long_s(const int32_t l);

  /// Consider @c p as a pointer to endian-swapped unsigned longs and
  ///  return *p.
  int32_t
  extract_swapped_long_s(const void* p);

  /// Return @c *p, possibly with endianness adjustment if @c swapped.
  int32_t
  extract_long_s(const void* p, bool swapped);

  /// Return @c s, possibly with endianness adjustment if @c swapped.
  int32_t
  extract_long_s(const int32_t l, bool swapped);

  /// Consider @c p as a pointer to big-endian unsigned longs and return *p.
  int32_t
  extract_big_endian_long_s(const void* p);

  /// Swap @c s if needed (i.e. host is little endian).
  int32_t
  extract_big_endian_long_s(const int32_t l);

  /// Consider @c p as a pointer to little-endian unsigned longs and
  /// return *p.
  int32_t
  extract_little_endian_long_s(const void* p);

  /// Swap @c s if needed (i.e. host is big endian).
  int32_t
  extract_little_endian_long_s(const int32_t l);

} // End of namespace tool.

# include "endianness.hxx"

#endif // ! TOOL_ENDIANNESS_HH_
