/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_OPTIONS_HXX_
# define TOOL_OPTIONS_HXX_

# include <typeinfo>
# ifdef __GNUC__
#  include <cxxabi.h>
# endif // __GNUC__
# include <vector>
# include <string>
# include <set>

# include <wipal/tool/options.hh>
# include <wipal/tool/endianness.hh>

namespace tool
{

  namespace opt
  {

    /*-------------.
    | invalid_type |
    `-------------*/

    inline
    invalid_type::invalid_type(const std::string& opt_type,
			       const std::string& req_type):
      std::logic_error ("(Internal error) option type is \"" + opt_type +
			"\", but \"" + req_type + "\" requested")
    {
    }

    /*------.
    | value |
    `------*/

    inline
    value::value()
    {
    }

    template <class T>
    value::value(const T& v): boost::any (v)
    {
    }

    template <class T>
    T&
    value::get()
    {
      try
	{
	  return boost::any_cast<T&>(*this);
	}
      catch (const boost::bad_any_cast&)
	{
	  throw invalid_type (type().name(), typeid(T).name());
	}
    }

    template <class T>
    const T&
    value::get() const
    {
      try
	{
	  return boost::any_cast<const T&>(*this);
	}
      catch (const boost::bad_any_cast&)
	{
	  throw invalid_type (type().name(), typeid(T).name());
	}
    }

    inline
    value&
    value::operator [] (const name& n)
    {
      return get<list>()[n];
    }

    inline
    const value&
    value::operator [] (const name& n) const
    {
      return get<list>()[n];
    }

    namespace internals
    {
      typedef std::vector<std::string>			vector_of_strings;
      typedef std::set< std::pair<unsigned, unsigned> >	set_of_uint_pairs;

      inline
      std::ostream&
      print_vector_of_strings(std::ostream& o, const vector_of_strings& v)
      {
	o << '[';
	for (vector_of_strings::const_iterator i = v.begin();
	     i != v.end();
	     ++i)
	  {
	    if (i != v.begin())
	      o << ", ";
	    o << '"' << *i << '"';
	  }
	return o << ']';
      }

      inline
      std::ostream&
      print_set_of_uint_pairs(std::ostream& o, const set_of_uint_pairs& s)
      {
	bool	empty (true);

	o << "{ ";
	for (set_of_uint_pairs::const_iterator i = s.begin();
	     i != s.end();
	     ++i)
	  {
	    if (not empty)
	      o << ", ";
	    o << i->first << '-' << i->second;
	    empty = false;
	  }
	return o << (empty? "}" : " }");
      }

    } // End of namespace tool::opt::internals.

    inline
    std::ostream&
    value::print(std::ostream& o) const
    {
      const std::type_info&	ti = type();

      if (false)
	;

# define WP_print_type(T)			\
      else if (ti == typeid (T))		\
	return o << get<T>()

      WP_print_type(bool);
      WP_print_type(int);
      WP_print_type(unsigned);
      WP_print_type(std::string);
      WP_print_type(list);
      WP_print_type(endian::endianness);

# undef WP_print_type

      else if (ti == typeid (internals::vector_of_strings))
	return internals::
	  print_vector_of_strings(o, get<internals::vector_of_strings>());
      else if (ti == typeid (internals::set_of_uint_pairs))
	return internals::
	  print_set_of_uint_pairs(o, get<internals::set_of_uint_pairs>());
      else
	return o << '['
# ifdef __GNUC__
		 << abi::__cxa_demangle(ti.name(), 0, 0, 0)
# else // ! __GNUC__
		 << ti.name()
#endif // __GNUC__
		 << ']';
    }

    inline
    std::ostream&
    operator << (std::ostream& o, const value& v)
    {
      return v.print(o);
    }

    template <class T>
    T&
    get(value& v)
    {
      return v.get<T>();
    }

    template <class T>
    const T&
    get(const value& v)
    {
      return v.get<T>();
    }

    /*-----.
    | name |
    `-----*/

    inline
    name::name(const std::string& n): std::string (n)
    {
    }

    inline
    name::name(const char* n): std::string (n)
    {
    }

    inline
    internals::pair
    name::operator -= (const value& v) const
    {
      return internals::pair (*this, v);
    }


    namespace internals
    {

      /*-----.
      | pair |
      `-----*/

      inline
      pair::pair(const name& n, const value& v): opt_name (n), opt_value (v)
      {
      }

      inline
      defaults_list<pair>
      pair::operator , (const pair& p) const
      {
	return defaults_list<pair> (p, *this);
      }

      template <class Map>
      void
      pair::populate(Map& m) const
      {
	m.insert(std::make_pair(opt_name, opt_value));
      }

      /*--------------.
      | defaults_list |
      `--------------*/

      template <class NDL>
      defaults_list<NDL>::defaults_list(const pair& p, const NDL& next):
	this_pair (p),
	next_defaults (next)
      {
      }

      template <class NDL>
      defaults_list< defaults_list<NDL> >
      defaults_list<NDL>::operator , (const pair& p) const
      {
	return defaults_list< defaults_list<NDL> > (p, *this);
      }

      template <class NDL>
      template <class Map>
      void
      defaults_list<NDL>::populate(Map& m) const
      {
	this_pair.populate(m);
	next_defaults.populate(m);
      }

    } // End of namespace tool::opt::internals.


    /*---------------.
    | invalid_option |
    `---------------*/

    inline
    invalid_option::invalid_option(const name& option):
      invalid_argument (option + " is not a valid option")
    {
    }


    /*-----.
    | list |
    `-----*/

    inline
    list::list()
    {
    }

    inline
    list::list(const internals::pair& defaults)
    {
      defaults.populate(*this);
    }

    template <class D>
    list::list(const internals::defaults_list<D>& defaults)
    {
      defaults.populate(*this);
    }

    inline
    value&
    list::operator [] (const name& n)
    {
      iterator	i = find(n);

      if (i == end())
	throw invalid_option (n);

      return i->second;
    }

    inline
    const value&
    list::operator [] (const name& n) const
    {
      const_iterator	i = find(n);

      if (i == end())
	throw invalid_option (n);

      return i->second;
    }

    inline
    void
    list::add(const list& l)
    {
      l.populate(*this);
    }

    inline
    void
    list::add(const name& n, const value& v)
    {
      if (v.type() == typeid (list))
	add(n, v.get<list>());
      else
	super_type::operator [] (n) = v;
    }

    inline
    list&
    list::add(const name& n, const list& l)
    {
      iterator	i = find(n);

      if (i == end())
	i = insert(std::make_pair(n, l)).first;
      else
	{
	  if (i->second.type() != typeid (list))
	    i->second = list ();
	  l.populate(get<list>(i->second));
	}
      return get<list>(i->second);
    }

    inline
    void
    list::populate(list& other) const
    {
      for (const_iterator i = begin(); i != end(); ++i)
	other.add(i->first, i->second);
    }

    inline
    std::ostream&
    list::print(std::ostream& os, unsigned depth) const
    {
      if (empty())
	return os << "<empty option list>" << std::endl;

      for (const_iterator i = begin(); i != end(); ++i)
	{
	  for (unsigned j = 0; j < depth; ++j)
	    os << "  ";
	  if (i->second.type() == typeid (list))
	    {
	      os << "- " << i->first << '\n';
	      get<list>(i->second).print(os, depth + 1);
	    }
	  else
	    os << i->first << ": " << i->second << '\n';
	}
      return os << std::flush;
    }

    inline
    std::ostream&
    operator << (std::ostream& os, const list& l)
    {
      return l.print(os);
    }

  } // End of namespace tool::opt.

} // End of namespace tool.

#endif // TOOL_OPTIONS_HXX_
