/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_ANONYMIZER_HH_
# define WIFI_FRAME_FILTER_ANONYMIZER_HH_

extern "C"
{
# include <stdint.h>
}

# include <set>

# include <wipal/tool/iterable.hh>
# include <wipal/tool/valued_iterator.hh>
# include <wipal/pcap/frame_descriptor.hh>
# include <wipal/wifi/addr.hh>
# include <wipal/wifi/frame/dissector/default_hooks.hh>
# include <wipal/wifi/essid.hh>
# include <wipal/wifi/frame/filter/anon_policies.hh>

/*---------------------.
| Forward declarations |
`---------------------*/

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      template <class, class, class, class>
      struct anonymizer;

      namespace internals
      {

	template <class, class, class, class, class>
	struct anonymizer_iterator;

      } // End of namespace wifi::frame::filter::internals.

    } // End of namespace wifi::frame::filter.
  } // End of namespace wifi::frame.
} // End of namespace wifi.


/*--------------.
| Virtual types |
`--------------*/

namespace tool
{

  template <class I, class P, class A, class B>
  struct types< wifi::frame::filter::anonymizer<I, P, A, B> >
  {
    typedef wifi::frame::filter::internals::
    anonymizer_iterator<I, P, A, B, tool::bottom>		iterator;
  };

  template <class I, class P, class A, class B1, class B2>
  struct types< wifi::frame::filter::internals::
                anonymizer_iterator<I, P, A, B1, B2> >
  {
    typedef wifi::frame::filter::anonymizer<I, P, A, B1>	iterable_type;
    typedef typename pcapxx::frame_descriptor<>			value_type;
  };

} // End of namespace tool.


/*-------------.
| Declarations |
`-------------*/

namespace wifi
{
  namespace frame
  {

    /// Anonymization policies.
    namespace anon
    {

      struct dummy_mac_policy
      {
	addr		operator [] (const addr& mac_addr) const;
      };

      struct dummy_essid_policy
      {
	essid		operator [] (const essid& essid) const;
      };

      template <class MACAnonPolicy, class ESSIDAnonPolicy>
      struct generic_policy: private MACAnonPolicy, private ESSIDAnonPolicy
      {
	generic_policy(const MACAnonPolicy& macp = MACAnonPolicy (),
		       const ESSIDAnonPolicy& essidp = ESSIDAnonPolicy ());

	addr		operator [] (const addr& mac_addr);
	essid		operator [] (const essid& essid);
      };

      typedef generic_policy<dummy_mac_policy, dummy_essid_policy>
	      dummy_policy;

      typedef generic_policy<prefix_preserving_mac_policy,
			     prefix_preserving_essid_policy>
	      prefix_preserving_policy;

      typedef prefix_preserving_policy
	      default_policy;

    } // End of namespace wifi::frame::anon.


    namespace filter
    {

      namespace internals
      {

	/**
	 * Iterator class of wifi::frame::filter::anonymizer.
	 *
	 * anonymizer_iterator::increment() is the anonymization
	 * process' core.
	 */
	template <class I, class P, class A, class B,
		  class Bottom = tool::bottom>
	struct anonymizer_iterator: WP_INHERIT(public tool::valued_iterator,
					       anonymizer_iterator<I, P, A, B,
								   Bottom>),
				    private A // Anonymization policy.
	{
	  typedef WP_GET_EXACT(Bottom, anonymizer_iterator<I, P, A, B, Bottom>)
							exact_type;
	  typedef WP_TYPE(value_type, exact_type)	value_type;
	  typedef WP_TYPE(iterable_type, exact_type)	iterable_type;

	  anonymizer_iterator(const iterable_type& iterable, bool end);

	  bool	equal(const anonymizer_iterator& rhs) const;
	  void	increment();

	private:
	  void	anonymize_addrs(const uint8_t* const		phy,
				const std::set<const addr*>&	addrs,
				uint8_t* const			anon_bytes,
				const size_t			anon_caplen);

	  size_t
	    anonymize_ssid(const typename I::value_type&	frm,
			   const ssid_elt* const		ssid,
			   uint8_t* const			anon_bytes,
			   const size_t				anon_caplen);

	  struct hooks: public dissector_default_hooks
	  {
	    hooks();

	    void
	    addr_hook(const void*, size_t, unsigned, const addr&);

	    void
	    end_of_frame_headers_hook(const void*, size_t, const void*);

	    type::frame_type		ty;
	    const uint8_t*		end;
	    std::set<const addr*>	addrs;
	    const ssid_elt*		ssid;
	  };

	  I			next_;
	  I			last_;
	};

      } // End of namespace wifi::frame::filter::internals.


      template <class InputIterator,
		class Phy,
		class AnonPolicy = anon::default_policy,
		class Bottom = tool::bottom>
      struct anonymizer: WP_INHERIT(public tool::iterable,
				    anonymizer<InputIterator, Phy, AnonPolicy,
					       Bottom>),
			 private AnonPolicy
      {
	anonymizer(const InputIterator& first, const InputIterator& last,
		   const AnonPolicy& anon_policy = AnonPolicy ());

      private:
	InputIterator	first_;
	InputIterator	last_;

	friend struct internals::anonymizer_iterator<InputIterator,
						     Phy,
						     AnonPolicy,
						     Bottom, tool::bottom>;
      };

    } // End of namespace wifi::frame::filter.
  } // End of namespace wifi::frame.
} // End of namespace wifi.

# include "anonymizer.hxx"

#endif // ! WIFI_FRAME_FILTER_ANONYMIZER_HH_
