/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_STATS_SIMPLE_COUNTERS_HXX_
# define WIFI_FRAME_STATS_SIMPLE_COUNTERS_HXX_

# include <cstring>
# include <cassert>

# include <wipal/wifi/frame/stats/simple_counters.hh>
# include <wipal/tool/si.hh>
# include <wipal/wifi/frame/frame.hh>

namespace wifi
{

  namespace frame
  {

    namespace stats
    {

      inline
      simple_counters::simple_counters(const void*): total_ (0),
						     size_ (0),
						     ap_total_ (0),
						     ap_size_ (0),
						     retransmitted_ (0),
						     expired_ (0),
						     out_of_order_ (0),
						     duplicate_ (0),
						     gap_ (0),
						     seq_anomaly_ (0)
      {
	memset(type_count_, 0, sizeof (type_count_));
	memset(type_size_, 0, sizeof (type_size_));
	memset(subtype_count_, 0, sizeof (subtype_count_));
	memset(subtype_size_, 0, sizeof (subtype_size_));
      }

      inline
      void
      simple_counters::account_duplicate()
      {
	++duplicate_;
      }

      inline
      void
      simple_counters::account_expired()
      {
	++expired_;
      }

      template <class Frame>
      void
      simple_counters::
      account_frame(const Frame&		,
		    const size_t&		len80211,
		    const type::frame_type&	type,
		    const unsigned&		subtype,
		    const bool&			retx,
		    const addr* const&		,
		    const addr* const&		,
		    const bool&			from_ap)
      {
	assert(type < 4);
	assert(subtype < 16);

	++total_;
	size_ += len80211;

	if (from_ap)
	  {
	    ++ap_total_;
	    ap_size_ += len80211;
	  }

	if (retx)
	  ++retransmitted_;

	++type_count_[type];
	type_size_[type] += len80211;

	++subtype_count_[type][subtype];
	subtype_size_[type][subtype] += len80211;
      }

      inline
      void
      simple_counters::account_gap(const unsigned&)
      {
	++gap_;
      }

      inline
      void
      simple_counters::account_out_of_order()
      {
	++out_of_order_;
      }

      inline
      void
      simple_counters::account_seq_anomaly()
      {
	++seq_anomaly_;
      }

      inline
      std::ostream&
      simple_counters::print(std::ostream& o) const
      {
	o << "frame count: " << total_ << '\n'
	  << "traffic size (B): " << size_ << '\n'
	  << "traffic size (MiB): " << tool::MiB (tool::B (size_)) << '\n'
	  << "AP frame count: " << ap_total_ << '\n'
	  << "AP traffic size: " << tool::MiB (tool::B (ap_size_)) << '\n'
	  << "AP traffic ratio:" << (double (ap_size_) / size_) << '\n'
	  << "non-AP frame count: " << (total_ - ap_total_) << '\n'
	  << "non-AP traffic size: "
	  <<                     tool::MiB (tool::B (size_ - ap_size_)) << '\n'
	  << "non-AP traffic ratio:"
	  <<                        (double (size_ - ap_size_) / size_) << '\n'
	  << "mean frame length: " << (double (size_) / total_) << '\n'
	  << "retransmissions: " << retransmitted_ << '\n'
	  << "retransmissions ratio: "
	  <<   (double (retransmitted_) / total_) << '\n'
	  << "frames from expired senders: " << expired_ << '\n'
	  << "frames from expired senders ratio: "
	  <<   (double (expired_) / total_) << '\n'
	  << "out-of-order frames: " << out_of_order_ << '\n'
	  << "out-of-order frames ratio: "
	  <<   (double (out_of_order_) / total_) << '\n'
	  << "duplicate frames w.r.t. sequence numbers: " << duplicate_ << '\n'
	  << "duplicate frames w.r.t. sequence numbers ratio: "
	  <<   (double (duplicate_) / total_) << '\n'
	  << "sequence gap too large to make sense: " <<   seq_anomaly_ << '\n'
	  << "sequence gap too large to make sense ratio (wrt. frame count): "
	  <<   (double (seq_anomaly_) / total_) << '\n'
	  << "sequence gap too large to make sense ratio (wrt. gap count): "
	  <<   (double (seq_anomaly_) / (seq_anomaly_ + gap_)) << '\n'
	  << '\n'
	  << "begin frame types/subtypes\n";
	for (unsigned t = 0; t < 4; ++t)
	  {
	    const unsigned	tc = type_count_[t];
	    const uint64_t	ts = type_size_[t];

	    if (not tc)
	      continue;

	    o << '\n'
	      << "  begin type: " << type::names[t] << '\n'
	      << "    count: " << tc << '\n'
	      << "    size (B): " << ts << '\n'
	      << "    size (MiB): " << tool::MiB (tool::B (ts)) << '\n'
	      << "    mean frame length: " << (double (ts) / tc) << '\n'
	      << "    count ratio: " << (double (tc) / total_) << '\n'
	      << "    traffic ratio: " << (double (ts) / size_) << '\n';
	    for (unsigned st = 0; st < 16; ++st)
	      {
		const unsigned	stc = subtype_count_[t][st];
		const uint64_t	sts = subtype_size_[t][st];

		if (not stc)
		  continue;

		o << '\n'
		  << "    begin subtype: "
		  <<          type::names[t] << '/'
		  <<          subtype::names[t][st] << '\n'
		  << "      count: " << stc << '\n'
		  << "      size (B): " << sts << '\n'
		  << "      size (MiB): " << tool::MiB (tool::B (sts)) << '\n'
		  << "      mean frame length: " << (double (sts) / stc) << '\n'
		  << "      count ratio (overall): "
		  <<          (double (stc) / total_) << '\n'
		  << "      count ratio (this type): "
		  <<          (double (stc) / tc) << '\n'
		  << "      traffic ratio (overall): "
		  <<          (double (sts) / size_) << '\n'
		  << "      traffic ratio (this type): "
		  <<          (double (sts) / ts) << '\n'
		  << "    end subtype: " << type::names[t] << '/'
		  <<                        subtype::names[t][st] << '\n';
	      }
	    o << "\n  end type: " << type::names[t] << '\n';
	  }
	o << "\nend frame types/subtypes" << std::endl;
	return o;
      }

      inline
      unsigned
      simple_counters::total() const
      {
	return total_;
      }

      inline
      unsigned
      simple_counters::type_count(unsigned type) const
      {
	return type_count_[type];
      }

      inline
      unsigned
      simple_counters::subtype_count(unsigned type, unsigned subtype) const
      {
	return subtype_count_[type][subtype];
      }

    } // End of namespace wifi::frame::stats.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_STATS_SIMPLE_COUNTERS_HXX_
