/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <iostream>
#include <stdexcept>
#include <cstdlib>

#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include <wipal/tool/args.hh>
#include <wipal/wifi/frame/filter/intersector.hh>
#include <wipal/wifi/frame/filter/fast_intersector.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/addr_mapping.hh>

namespace
{

  struct intersection_dumper
  {

    intersection_dumper(const tool::opt::list& options):
      microseconds_ (options["ui"]["microseconds"].get<bool>())
    {
    }

    template <class Intersector>
    void	operator () (const Intersector& i)
    {
      typedef typename Intersector::const_iterator	iterator;

      const iterator		end = i.end();

      for (iterator j = i.begin(); j != end; ++j)
	{
	  if (microseconds_)
	    std::cout << j->first.microseconds() << '\t';
	  std::cout << j->first << '\t';
	  if (microseconds_)
	    std::cout << j->second.microseconds() << '\t';
	  std::cout << j->second << '\n';
	}
    }

  private:
    const bool&	microseconds_;
  };

  struct main_helper
  {
    template <class UniqueId>
    int
    operator () (tool::opt::list& options) const
    {
      namespace opt = tool::opt;
      namespace args = tool::args;

      using wifi::frame::filter::provide_intersector;

      const args::input&	input = opt::get<args::input>(options["proc"]
							      ["input"]);
      pcapxx::list_of_traces<>	traces1 (input[0]);
      pcapxx::list_of_traces<>	traces2 (input[1]);
      intersection_dumper	dump (options);

      if (traces1.empty() or traces2.empty())
	return EXIT_SUCCESS;

      const pcapxx::descriptor<>::link_type	l = traces1.get(0).linktype();

      args::check_options(options, l);
      traces1.expect(l);
      traces2.expect(l);

      switch (l)
	{

#define WP_case(R, Unused, LT)						\
	case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	  provide_intersector<UniqueId, BOOST_PP_TUPLE_ELEM(2, 1, LT),	\
	    wifi::frame::filter::fast_intersector>			\
	    (traces1.begin(), traces1.end(),				\
	     traces2.begin(), traces2.end(),				\
	     opt::get<opt::list>(options["internals"]), dump);		\
	  break;

	  BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

#undef WP_case

	default:
	  args::unsupported_linktype(l);
	}
      return EXIT_SUCCESS;
    }

  }; // End of struct main_helper.

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      namespace opt = tool::opt;
      namespace args = tool::args;
      namespace def = tool::args::def;

      opt::list opts ((opt::name ("internals") -= opt::list
		       ((opt::name ("endianness") -= def::endianness(),
			 opt::name ("ignore_noisy_prism") -=
						     def::ignore_noisy_prism(),
			 opt::name ("mapping") -= def::mapping(),
			 opt::name ("precision") -= def::precision(),
			 opt::name ("phy_tstamps") -= def::phy_tstamps())),

		       opt::name ("misc") -= opt::list
		       ((opt::name ("mac_map") -= def::mac_map(),
			 opt::name ("attributes") -= def::attributes())),

		       opt::name ("ui") -= opt::list
		       ((opt::name ("microseconds") -=
						    def::ui_microseconds()))));

      args::parse(argc, argv, opts, args::count (2), args::none);
      return args::dispatch_unique_id(opts, main_helper ());
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
