/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <stdexcept>
#include <vector>
#include <cstdlib>

#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include <wipal/tool/args.hh>
#include <wipal/wifi/traces/similarity.hh>
#include <wipal/wifi/frame/filter/fast_intersector.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/addr_mapping.hh>

namespace
{

  struct main_helper
  {
    template <class UniqueId>
    int
    operator () (tool::opt::list& options) const
    {
      namespace opt = tool::opt;
      namespace args = tool::args;

      using wifi::frame::filter::fast_intersector;

      typedef pcapxx::list_of_traces<>	trace_type;
      typedef std::vector<trace_type>	trace_vector_type;
      typedef boost::tuple<unsigned, trace_vector_type::iterator,
				     trace_vector_type::iterator>
					score_type;
      typedef std::set<score_type>	score_set_type;

      trace_vector_type			traces;
      score_set_type			scores;
      pcapxx::descriptor<>::link_type	ltype;
      bool				ltype_ok = false;

      BOOST_FOREACH(trace_type t,
		    opt::get<args::input>(options["proc"]["input"]))
	{
	  if (not ltype_ok)
	    {
	      if (t.empty())
		throw std::invalid_argument (t.file_name() + ": empty trace");
	      ltype = t.get(0).linktype();
	      ltype_ok = true;
	      args::check_options(options, ltype);
	    }
	  t.expect(ltype);
	  traces.push_back(t);
	}

      if (traces.empty())
	return EXIT_SUCCESS;

      switch (ltype)
	{

#define WP_case(R, Unused, LT)						\
	case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	  scores = wifi::traces::similarity<UniqueId,			\
					    BOOST_PP_TUPLE_ELEM(2, 1, LT), \
					    fast_intersector>		\
	    (traces, options);						\
	  break;

	  BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

#undef WP_case

	default:
	  args::unsupported_linktype(ltype);
	}

      BOOST_FOREACH(const score_type& s, scores)
	std::cout << s.get<0>() << '\t'
		  << s.get<1>()->file_name() << '\t'
		  << s.get<2>()->file_name() << '\n';
      return EXIT_SUCCESS;
    }

  }; // End of struct main_helper.

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      namespace opt = tool::opt;
      namespace args = tool::args;
      namespace def = tool::args::def;

      opt::list opts ((opt::name ("internals") -= opt::list
		       ((opt::name ("endianness") -= def::endianness(),
			 opt::name ("ignore_noisy_prism") -=
						     def::ignore_noisy_prism(),
			 opt::name ("mapping") -= def::mapping(),
			 opt::name ("precision") -= def::precision(),
			 opt::name ("phy_tstamps") -= def::phy_tstamps())),

		       opt::name ("misc") -= opt::list
		       ((opt::name ("mac_map") -= def::mac_map(),
			 opt::name ("attributes") -= def::attributes()))));

      args::parse(argc, argv, opts, args::arbitrary, args::none);
      return args::dispatch_unique_id(opts, main_helper ());
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
