/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <iostream>
#include <stdexcept>
#include <cstdlib>

#include <boost/format.hpp>
#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include <gmpxx.h>

#include <wipal/tool/args.hh>
#include <wipal/wifi/frame/filter/linear_regression_synchronizer.hh>
#include <wipal/wifi/frame/filter/intersector.hh>
#include <wipal/wifi/frame/filter/fast_intersector.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/addr_mapping.hh>

namespace
{

  struct sync_unique_frames_dumper
  {

    sync_unique_frames_dumper(const tool::opt::list& options):
      print_headers_ (options["ui"]["headers"].get<bool>()),
      print_indices_ (options["ui"]["indices"].get<bool>()),
      print_msecs_ (options["ui"]["microseconds"].get<bool>()),
      print_errors_ (options["ui"]["errors"].get<bool>()),
      coefs_ ("%10s %15s"),
      indices_ (" %10s %10s"),
      msecs_   (" %15s %15s"),
      errors_  (" %10s")
    {
    }

    void	print_colheads()
    {
      std::cout << (coefs_ % "a" % "b");
      if (print_indices_)
	std::cout << (indices_ % "frame 1" % "frame 2");
      if (print_msecs_)
	std::cout << (msecs_ % "time 1 (us)" % "time 2 (us)");
      if (print_errors_)
	std::cout << (errors_ % "error");

      std::cout << '\n'
		<< (coefs_ % "=" % "=");
      if (print_indices_)
	std::cout << (indices_ % "=======" % "=======");
      if (print_msecs_)
	std::cout << (msecs_ % "===========" % "===========");
      if (print_errors_)
	std::cout << (errors_ % "=====");
      std::cout << std::endl;
    }

    template <class LRSync>
    void	operator () (const LRSync& s)
    {
      typedef typename LRSync::const_iterator	iterator;

      if (print_headers_)
	print_colheads();

      const iterator	end = s.end();

      for (iterator j = s.begin(); j != end; ++j)
	{
	  std::cout << (coefs_ % j->coefs.first % j->coefs.second);
	  if (print_indices_)
	    std::cout << (indices_ % j->first % j->second);
	  if (print_msecs_)
	    std::cout << (msecs_ % j->first.microseconds()
			         % j->second.microseconds());
	  if (print_errors_)
	    std::cout << (errors_ % (j->first.microseconds() * j->coefs.first +
				     j->coefs.second -
				     j->second.microseconds()));
	  std::cout << '\n';
	}
    }

  private:
    const bool&		print_headers_;
    const bool&		print_indices_;
    const bool&		print_msecs_;
    const bool&		print_errors_;
    boost::format	coefs_;
    boost::format	indices_;
    boost::format	msecs_;
    boost::format	errors_;
  };

  struct main_helper
  {
    template <class UniqueId>
    int
    operator () (tool::opt::list& options) const
    {
      namespace opt = tool::opt;
      namespace args = tool::args;

      using wifi::frame::filter::provide_lr_synchronizer;

      mpf_set_default_prec(WP_GMP_FLOAT_PRECISION);

      const args::input&	input = opt::get<args::input>(options["proc"]
							      ["input"]);
      pcapxx::list_of_traces<>	traces1 (input[0]);
      pcapxx::list_of_traces<>	traces2 (input[1]);
      sync_unique_frames_dumper	dump (options);

      if (traces1.empty() or traces2.empty())
	return EXIT_SUCCESS;

      const pcapxx::descriptor<>::link_type	l = traces1.get(0).linktype();

      args::check_options(options, l);
      traces1.expect(l);
      traces2.expect(l);

      switch (l)
	{

#define WP_case(R, Unused, LT)						\
	case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	  provide_lr_synchronizer<UniqueId, BOOST_PP_TUPLE_ELEM(2, 1, LT), \
	    wifi::frame::filter::fast_intersector>			\
	    (traces1.begin(), traces1.end(),				\
	     traces2.begin(), traces2.end(),				\
	     opt::get<opt::list>(options["internals"]), dump);		\
	  break;

	  BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

#undef WP_case

	default:
	  args::unsupported_linktype(l);
	}
      return EXIT_SUCCESS;
    }

  }; // End of struct main_helper

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      namespace opt = tool::opt;
      namespace args = tool::args;
      namespace def = tool::args::def;

      opt::list	opts ((opt::name ("internals") -= opt::list
		       ((opt::name ("endianness") -= def::endianness(),
			 opt::name ("ref_blacklist") -= def::ref_blacklist(),
			 opt::name ("ignore_noisy_prism") -=
						     def::ignore_noisy_prism(),
			 opt::name ("mapping") -= def::mapping(),
			 opt::name ("precision") -= def::precision(),
			 opt::name ("phy_tstamps") -= def::phy_tstamps())),

		       opt::name ("misc") -= opt::list
		       ((opt::name ("mac_map") -= def::mac_map(),
			 opt::name ("attributes") -= def::attributes())),

		       opt::name ("ui") -= opt::list
		       ((opt::name ("headers") -= def::ui_headers(),
			 opt::name ("errors") -= def::ui_errors(),
			 opt::name ("indices") -= def::ui_indices(),
			 opt::name ("microseconds") -=
						    def::ui_microseconds()))));

      args::parse(argc, argv, opts, args::count (2), args::none);
      return args::dispatch_unique_id(opts, main_helper ());
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
